/***
 * Portable GUI
 * Code for HelloWorld
 * Copyright (C) 2000-2000 Jan Struyf <jan.struyf@cs.kuleuven.ac.be>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
 
// Import all PortableGUI include files
#include <portablegui.h>

// For getenv
#ifdef PGGTK
	#include <stdlib.h>
#endif

// Blocksize for load/save operations
#define BLOCKSIZE 4000

// Preferences for this application
class MyPreferences : public PGObject {
public:	
	int m_iTitleSize;	
	BOOL m_bStatusBar;
	BOOL m_bSaveOnExit;
	BOOL m_bOldSaveOnExit;
	BOOL m_bModified;
public:
	MyPreferences();
	void loadPreferences();	
	void saveOnExit();	
	void getINIFileName(PGString* fname);
};

// Text field for editing
PGTextArea* g_hTextArea;
// Main frame window
PGFrame* g_hMainFrame;
// Status bar
PGPanel* g_hStatus;
// Filter for filenames
char* g_sFilter = "Text files (*.txt)$*.txt$All files (*.*)$*.*$";
// File already named?
BOOL g_bHasName = false;
PGString g_sFile;
// Image for about box
PGImage* g_hImageAbout;
// Icon for Win/GNOME taskbar
PGIcon* g_hIcon;
// EditPad Preferences
MyPreferences g_cPreferences;

// Forward declarations for new, open and save methods
BOOL doNew();
BOOL doOpen();
BOOL doSave(BOOL verifyOverwrite);

// Update the window title bar
void UpdateTitle();
// Enable/Disable status bar
void UpdateStatus();

// Called when user tries to close the main window
class MyWindowListener : public PGWindowListener {
	
	// Exit the application	
	virtual void windowClosing(PGWindow* wnd) {		
		g_cPreferences.saveOnExit();
		getPGApplication()->exit(0); 
	}
};

// Called when user selects File | Exit
class MyExitListener : public PGActionListener {

	// Exit the application
	virtual void actionPerformed(PGActionListenerParent* from) {
		g_cPreferences.saveOnExit();
		getPGApplication()->exit(0); 
	}
};

// Called when user selects Edit | Cut
class MyMenuClosingListener : public PGMenuClosingListener {

	virtual void menuClosing(PGMenuBar* bar) {		
		g_hMainFrame->setStatus("EditPad");
	}
};


// Called when user selects File | New
class MyNewListener : public PGActionListener {

	// Call doNew and update windows title
	virtual void actionPerformed(PGActionListenerParent* from) {
		if (doNew()) g_bHasName = false;
		UpdateTitle();
	}
};

// Called when user selects File | Open
class MyOpenListener : public PGActionListener {

	// Open file
	virtual void actionPerformed(PGActionListenerParent* from) {
		// Create file open dialog
		PGFileDialog dial("Open");
		dial.setFilter(g_sFilter);
		if (g_bHasName) dial.setFile(g_sFile);
		// Display file open dialog
		if (dial.showFileOpenDialog(g_hMainFrame)) {
			g_sFile.setCopyValue(dial.m_sFile);
			g_bHasName = doOpen();
			UpdateTitle();
		}
	}
};

// Called when user selects File | Save
class MySaveListener : public PGActionListener {

	virtual void actionPerformed(PGActionListenerParent* from) {
		if (!g_bHasName) {
			// Create file dialog if file has no name yet
			PGFileDialog dial("Save");
			dial.setFilter(g_sFilter);
			// Display file dialog
			if (dial.showFileSaveDialog(g_hMainFrame)) {
				g_sFile.setCopyValue(dial.m_sFile);
			} else {
				return;
			}
		}
		// Save file and update window title
		g_bHasName = doSave(false);
		UpdateTitle();
	}
};

// Called when user selects File | Save As
class MySaveAsListener : public PGActionListener {

	virtual void actionPerformed(PGActionListenerParent* from) {
		// Create save as file dialog
		PGFileDialog dial("Save As");
		dial.setFilter(g_sFilter);
		if (g_bHasName) dial.setFile(g_sFile);
		// Display dialog, save and update title
		if (dial.showFileSaveDialog(g_hMainFrame)) {
			g_sFile.setCopyValue(dial.m_sFile);
			g_bHasName = doSave(true);
			UpdateTitle();
		}
	}
};

// Does GtkText support undo?
#ifdef PGWIN
	// Called when user selects File | Undo
	class MyUndoListener : public PGActionListener {

		virtual void actionPerformed(PGActionListenerParent* from) {
			g_hTextArea->undoOperation();
		}
	};
#endif

// Called when user selects Edit | Cut
class MyCutListener : public PGActionListener {

	virtual void actionPerformed(PGActionListenerParent* from) {
		g_hTextArea->cut();
	}
};

// Called when user selects Edit | Copy
class MyCopyListener : public PGActionListener {

	virtual void actionPerformed(PGActionListenerParent* from) {
		g_hTextArea->copy();
	}
};

// Called when user selects Edit | Paste
class MyPasteListener : public PGActionListener {

	virtual void actionPerformed(PGActionListenerParent* from) {
		g_hTextArea->paste();
	}
};

// Called when user selects Edit | Select All
class MySelectAllListener : public PGActionListener {

	virtual void actionPerformed(PGActionListenerParent* from) {
		g_hTextArea->selectAll();
	}
};

// Class representing the preference dialog
class MyPreferenceDialog : public PGDialog {
public:
	PGCheckbox* m_hStatusBarBox;
	PGCheckbox* m_hSaveOnExitBox;
	PGIntegerField* m_hTitleSizeField;
public:
	MyPreferenceDialog(PGWindow* parent);
	void onOKButton();
};

// Called when user clicks PreferenceDialog | OK
class MyPreferenceOKListener : public PGActionListener {
private:
	MyPreferenceDialog* m_hDialog;
public:	
	MyPreferenceOKListener(MyPreferenceDialog* dial) {
		m_hDialog = dial;
	}

	virtual void actionPerformed(PGActionListenerParent* from) {		
		m_hDialog->onOKButton();
		m_hDialog->dispose();
		UpdateTitle();
		UpdateStatus();
	}
};

// Create the preference dialog
MyPreferenceDialog::MyPreferenceDialog(PGWindow* parent) 
				: PGDialog(parent, "EditPad Preferences") {
	setIconImage(g_hIcon);
	addWindowListener(new PGCloseWindowListener());
	/***
	 * Percentlayout
	 * Component 1 : pl   = Preferred size, Left justify
	 *           2 : 20d  = Dummy component size 20 pixels
	 *           3 : 100% = 100% Stretch of leftover space
	 *           5 : pc   = Preferred size, Center justify
	 */
	setLayout(new PGPercentLayout("pl pl 10d p 100%d 20d p", 3, PGPERCENTLAYOUT_ALL, true));
	// Create status bar check box
	m_hStatusBarBox = new PGCheckbox("&Status bar");		
	m_hStatusBarBox->setToolTipText("Enable/Disable the status bar");
	m_hStatusBarBox->setState(g_cPreferences.m_bStatusBar);
	add(m_hStatusBarBox);
	// Create save preferences check box
	m_hSaveOnExitBox = new PGCheckbox("S&ave preferences on exit");		
	m_hSaveOnExitBox->setToolTipText("Save preferences every time EditPad is closed");
	m_hSaveOnExitBox->setState(g_cPreferences.m_bSaveOnExit);
	add(m_hSaveOnExitBox);
	// Create panel with entries
	PGPanel *entries = new PGPanel();
	entries->setLayout(new PGPercentLayout("p 100% p", 3, 0, false));
	entries->add(new PGLabel("Max filename length: "));
	m_hTitleSizeField = new PGIntegerField(g_cPreferences.m_iTitleSize);
	m_hTitleSizeField->setToolTipText("Set the maximum file name length for the title bar");
	entries->add(m_hTitleSizeField);		
	entries->add(new PGUpDown(m_hTitleSizeField, 5, 50, g_cPreferences.m_iTitleSize));
	add(entries);
	// Create panel with buttons
	PGPanel* panel = new PGPanel();
	panel->setLayout(new PGPercentLayout("p 100%d p", 3, 0, false));
	// Create and add OK button
	PGButton* ok = new PGButton(" &OK ", true);
	ok->setToolTipText("Close preference dialog and apply changes");
	ok->addActionListener(new MyPreferenceOKListener(this));
	panel->add(ok);
	// Create and add Cancel button
	PGButton* cancel = new PGButton(" &Cancel ", true);
	cancel->setToolTipText("Close preference dialog, discart changes");		
	cancel->addActionListener(new PGDialogCancelListener(this));
	panel->add(cancel);						
	// Pack components in window and show modal				
	add(panel);		
	pack();
}

// Called when user clicks PreferenceDialog | OK
void MyPreferenceDialog::onOKButton() {
	g_cPreferences.m_bStatusBar = m_hStatusBarBox->getState();
	g_cPreferences.m_bSaveOnExit = m_hSaveOnExitBox->getState();
	g_cPreferences.m_iTitleSize = m_hTitleSizeField->getValue();
	g_cPreferences.m_bModified = true;	
}

// Called when user selects Edit | Preferences
class MyPreferenceListener : public PGActionListener {

	// Create preferences dialog and do modal
	virtual void actionPerformed(PGActionListenerParent* from) {				
		MyPreferenceDialog pref(g_hMainFrame);				
		pref.doModal();		
	}
};

// Initialize EditPad preferences
MyPreferences::MyPreferences() {
	m_iTitleSize = 20;	
	m_bStatusBar = true;
	m_bSaveOnExit = false;
	m_bOldSaveOnExit = false;
	m_bModified = false;
}

// Find out name for INI file
void MyPreferences::getINIFileName(PGString* fname) {
#ifdef PGWIN
	fname->setCopyValue("EditPad.INI");
	fname->prepend(getPGApplication()->getStartDirectory());	
#endif
#ifdef PGGTK
	fname->setCopyValue("/.editpadrc");
	fname->prepend(getenv("HOME"));
#endif
}

// Load EditPad preferences from disk
void MyPreferences::loadPreferences() {
	PGString fname;
	getINIFileName(&fname);
	PGFileInputStream input(fname);
	if (input.getError() == PGINPUTSTREAM_ERROR_NONE) {
		PGINIFile inifile;
		inifile.read(&input);
		PGINIFileSection* section = inifile.getSection("EditPad");
		if (section != null) {
			PGString* value;
			value = section->getStringValue("StatusBar");
			if (value != null) m_bStatusBar = value->getBoolValue();
			value = section->getStringValue("SaveOnExit");
			if (value != null) m_bSaveOnExit = value->getBoolValue();
			value = section->getStringValue("TitleSize");
			if (value != null) m_iTitleSize = value->getIntValue();
		}
		m_bOldSaveOnExit = m_bSaveOnExit;
		m_bModified = false;		
	} else {
		printf("Error loading file: %s\n", fname.getValue());
	}
}

// Save EditPad preferences to disk
void MyPreferences::saveOnExit() {
	if (m_bModified && (m_bOldSaveOnExit || m_bSaveOnExit)) {
		PGString fname;
		getINIFileName(&fname);		
		PGFileOutputStream output(fname);
		if (output.getError() == PGOUTPUTSTREAM_ERROR_NONE) {
			PGINIFileWriter writer(&output);
			writer.startSection("EditPad");
			writer.writeBoolValue("StatusBar", m_bStatusBar);
			writer.writeBoolValue("SaveOnExit", m_bSaveOnExit);			
			writer.writeIntValue("TitleSize", m_iTitleSize);
			output.close();
		}
	}	
}

// Called when user selects Help | About
class MyAboutListener : public PGActionListener {

	virtual void actionPerformed(PGActionListenerParent* from) {
		// Create a new style (used for the about label)
		PGStyle* style = new PGStyle();
		style->setFont(new PGFont("Dialog", PGFONT_BOLD | PGFONT_ITALIC, 16));		
		// Create about dialog
		PGDialog about(g_hMainFrame, "About EditPad");
		about.setIconImage(g_hIcon);
		about.addWindowListener(new PGCloseWindowListener());		
		about.setLayout(new PGPercentLayout("pl 5d 100% 5d pl pl 10d pc", 3, PGPERCENTLAYOUT_ALL, true));
		// Create about label with large font style style
		PGLabel* label = new PGLabel("PortableGUI - EditPad demo");		
		label->setStyle(style);				
		about.add(label);
		// Create image viewer and some other labels		
		PGPanel* img = new PGPanel();
		img->setLayout(new PGPercentLayout("50%d p 50%d", 3, 0, false));
		img->add(new PGImageViewer(g_hImageAbout));
		about.add(img);		
		about.add(new PGLabel("Written by Jan Struyf"));
		about.add(new PGLabel("jan.struyf@cs.kuleuven.ac.be"));
		// Create and add close button
		PGButton* ok = new PGButton(" &Close ", true);
		ok->setToolTipText("Close this about window");
		ok->addActionListener(new PGDialogCancelListener(&about));
		about.add(ok);
		// Pack components in window and show modal		
		about.pack();		
		about.doModal();
	}
};

// Update window title bar
void UpdateTitle() {
	PGString myTitle;
	if (g_bHasName) {
		myTitle.setCopyValue(g_sFile);
		// Check if filename too long
		int len = myTitle.getLength();
		int max = g_cPreferences.m_iTitleSize;
		if (len > max) {		
			myTitle.substring(len-max, len);
			myTitle.prepend("EditPad - ..");
		} else {		
			myTitle.prepend("EditPad - ");
		}
	} else {
		myTitle.setCopyValue("EditPad");
	}
	// Set the filename in the title bar
	g_hMainFrame->setTitle(myTitle);
}

// Hide/Show the status bar
void UpdateStatus() {
	if (g_cPreferences.m_bStatusBar) {
		g_hMainFrame->add(g_hStatus, PGBORDERLAYOUT_SOUTH);		
		g_hStatus->setVisible(true);
		g_hMainFrame->updateLayout();		
	} else {
		g_hMainFrame->remove(g_hStatus);		
		g_hMainFrame->updateLayout();
	}	
}

// Clear text component
BOOL doNew() {
	g_hTextArea->setText("");
	return true;
}

// Save file to disk
BOOL doSave(BOOL verifyOverwrite) {	
	if (verifyOverwrite) {
		// Check if file exists
		PGFileInputStream input(g_sFile);				
		if (input.getError() == PGINPUTSTREAM_ERROR_NONE) {
			// File already exists, show message box
			input.close();
			PGMessageBox box("Save file","File already exists, overwrite?",PGMSGBOX_YES | PGMSGBOX_NO | PGMSGBOX_ICONQUESTION);
			if (box.show(g_hMainFrame) != PGMSGBOX_YES) return false;
		}
	}
	// Try to open file for output
	PGFileOutputStream output(g_sFile);
	if (output.getError() != PGOUTPUTSTREAM_ERROR_NONE) {
		// Can't open file, show message box
		PGMessageBox box("Save file","Can't create file",PGMSGBOX_OK | PGMSGBOX_ICONERROR);
		box.show(g_hMainFrame);
		return false;
	}
	// Copy contents of textarea to file
	PGString str;
	long tot = 0;
	// Don't update text component each block (works in GTK only)
	g_hTextArea->setDirectUpdate(false);
	long len = g_hTextArea->getSize();
	// Block max size is BLOCKSIZE
	do {
		// Calculate size for next block
		long now = MIN2(BLOCKSIZE, len-tot);
		// Get text block from text area and write to disk
		g_hTextArea->getText(&str, tot, tot+now);		
		int nb = output.write((unsigned char*)str.getValue(), 0, now);
		// Maybe out of quota? disk full?
		if (nb != (int)now) {
			g_hTextArea->setDirectUpdate(true);			
			output.close();
			PGMessageBox box("Save file","Disk full",PGMSGBOX_OK | PGMSGBOX_ICONERROR);
			box.show(g_hMainFrame);
			return false;
		}
		tot += now;
	} while (tot < len);
	g_hTextArea->setDirectUpdate(true);
	// Close file and return true
	output.close();
	return true;
}

// Load file from disk
BOOL doOpen() {
	if (!doNew()) return false;
	// Try to open file
	PGFileInputStream input(g_sFile);	
	if (input.getError() != PGINPUTSTREAM_ERROR_NONE) {
		// Can't open file, display message box
		PGMessageBox box("Open file","File not found",PGMSGBOX_OK | PGMSGBOX_ICONERROR);
		box.show(g_hMainFrame);
		return false;
	}
	long tot = 0;
	size_t nb = 0;
	// Don't update text component each block (works in GTK only)	
	g_hTextArea->setDirectUpdate(false);
	PGString str;
	str.reset(BLOCKSIZE);
	do {
		// Read in next block
		nb = input.read((unsigned char*)str.getValue(), 0, BLOCKSIZE);
		tot += nb;
		// Append block to text area
		g_hTextArea->append(str.getValue(), false);
	} while (nb == BLOCKSIZE);
	g_hTextArea->setDirectUpdate(true);
	input.close();	
#ifdef PGWIN
	// Maximum size for edit in windows
	// So, for a serious edit app, you have to write your own edit component	
	if (tot > 32000) {
		PGMessageBox box("Open file","File too large for editor",PGMSGBOX_OK | PGMSGBOX_ICONERROR);
		box.show(g_hMainFrame);
		return false;
	}
#endif
	return true;
}

// Create the main menu
PGMenuBar* makeMenu() {
	PGMenuBar* bar= new PGMenuBar();
	// Create file menu
	PGMenu* menu = new PGMenu("&File");
	PGMenuItem* item = new PGMenuItem("&New", "Create a new file");
	item->addActionListener(new MyNewListener());
	menu->add(item);
	item = new PGMenuItem("&Open", "Open a file from disk");
	item->addActionListener(new MyOpenListener());
	menu->add(item);
	item = new PGMenuItem("&Close", "Close the active file");
	menu->add(item);
	menu->add(new PGMenuSeparator());
	item = new PGMenuItem("&Save", "Save the active file to disk");
	item->addActionListener(new MySaveListener());
	menu->add(item);
	item = new PGMenuItem("Save &As", "Save the active file to disk with a new name");
	item->addActionListener(new MySaveAsListener());
	menu->add(item);
	menu->add(new PGMenuSeparator());
	item = new PGMenuItem("&Exit", "Close the EditPad application");
	item->addActionListener(new MyExitListener());
	menu->add(item);
	bar->add(menu);
	// Create edit menu
	menu = new PGMenu("&Edit");
#ifdef PGWIN
	item = new PGMenuItem("&Undo", "Undo last edit operation");
	item->addActionListener(new MyUndoListener());
	menu->add(item);
	menu->add(new PGMenuSeparator());
#endif
	item = new PGMenuItem("Cu&t", "Cut the selection and copy it to the clipboard");
	item->addActionListener(new MyCutListener());
	menu->add(item);
	item = new PGMenuItem("&Copy", "Copy the selection and put it on the clipboard");
	item->addActionListener(new MyCopyListener());
	menu->add(item);
	item = new PGMenuItem("&Paste", "Insert clipboard contents");
	item->addActionListener(new MyPasteListener());
	menu->add(item);
	item = new PGMenuItem("Select A&ll", "Select the entire file");
	item->addActionListener(new MySelectAllListener());
	menu->add(item);
	menu->add(new PGMenuSeparator());
	item = new PGMenuItem("&Preferences", "Change preferences");
	item->addActionListener(new MyPreferenceListener());
	menu->add(item);	
	bar->add(menu);
	// Create help menu
	menu = new PGMenu("&Help");
	item = new PGMenuItem("&About", "Show information about EditPad");
	item->addActionListener(new MyAboutListener());
	menu->add(item);
	bar->add(menu);
	bar->addMenuClosingListener(new MyMenuClosingListener());
	return bar;
}

PGPanel* makeStatus(PGFrame* frame, PGImage* file, BOOL enable) {
	// Create style for status line
	PGStyle* style = new PGStyle();
	style->setFont(new PGFont("Dialog", PGFONT_PLAIN, 8));
	// Create panel
	PGPanel* panel = new PGPanel();
	panel->setLayout(new PGPercentLayout("100% p", 3, 0, false));	
	// Create and add status line
	PGLabel* status = new PGLabel("EditPad", PGLABEL_SUNKEN);	
	status->setStyle(style);	
	panel->add(status);	
	// Add icon
	panel->add(new PGImageViewer(file));
	// Add panel to main frame
	if (enable) frame->add(panel, PGBORDERLAYOUT_SOUTH);
	frame->setStatusDisplay(status);		
	return panel;
}


// Mainproc is different for Windows and GTK
PGMAINPROC { 
	PGApplication app;	
	// Initialize the app using Mainproc args
	PGINIT(app);
	app.setConsoleEnabled(true); 
	// Load preferences
	g_cPreferences.loadPreferences();
	// Create main frame window
	g_hMainFrame = new PGFrame("EditPad");
	PGINCREF(g_hMainFrame);
	// Load about pixmap (BMP fileformat 4 or 8 bits/pixel)	
	g_hImageAbout = new PGImage("../about.bmp", g_hMainFrame);
	PGINCREF(g_hImageAbout);
	// Load icon for main window	
	PGColor background(178,178,180);	
	PGImage* file = new PGImage("../../images/file.bmp", &background, g_hMainFrame);	 
	g_hIcon = new PGIcon(file);
	g_hMainFrame->setIconImage(g_hIcon);
	// Add window listener to frame
	g_hMainFrame->addWindowListener(new MyWindowListener()); 
	// Use simple layoutmanager for the edit control
	g_hMainFrame->setLayout(new PGBorderLayout(1,1,3,3));
	// Create main menu and add menu bar
	g_hMainFrame->setMenuBar(makeMenu());
	// Add text area for editing
	g_hTextArea = new PGTextArea("", 30, 40, PGTEXTAREA_SCROLLBARS_VERTICAL_ONLY);	
	PGINCREF(g_hTextArea);
	g_hMainFrame->add(g_hTextArea, PGBORDERLAYOUT_CENTER);
	// Make and add status bar
	PGImage* resize = new PGImage("../../images/resize.bmp", &background, g_hMainFrame);
	g_hStatus = makeStatus(g_hMainFrame, resize, g_cPreferences.m_bStatusBar);
	PGINCREF(g_hStatus);
	// Pack and show the frame
	g_hMainFrame->pack(); 
	g_hTextArea->requestFocus();
	g_hMainFrame->setVisible(true);
	// Run the main event loop
	PGEXITCODE exitCode = app.run();
	// Free global references
	PGDECREF(g_hMainFrame);
	PGDECREF(g_hImageAbout);
	PGDECREF(g_hTextArea);
	PGDECREF(g_hStatus);
	return exitCode;
}
