/***
 * Portable GUI
 * Code for PGImage
 * Copyright (C) 2000-2000 Jan Struyf <jan.struyf@cs.kuleuven.ac.be>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef _PGIMAGE_
#define _PGIMAGE_

#include <pgmain/pgmain.h> 
#include <pgdraw/pgcolor.h>
#include <pgwindow/pgwindow.h>

#define PGIMAGE_ERROR_NONE	0	// Image has no error
#define PGIMAGE_ERROR_FILE	1	// Error opening image file
#define PGIMAGE_ERROR_CREATE	2	// Error creating image
#define PGIMAGE_ERROR_HEADER	3	// Error in image file header
#define PGIMAGE_ERROR_DATA	4	// Error decompressing image file data
#define PGIMAGE_ERROR_MEMORY	5	// Error allocating memory
#define PGIMAGE_ERROR_NOCREATE	6	// Image has not yet been created
#define PGIMAGE_ERROR_UNSUPPORT 7
    
#define PG_DISPOSAL_NONE		0
#define PG_DISPOSAL_LEAVE_IMAGE		1
#define PG_DISPOSAL_RESTORE_BACKGROUND	2
#define PG_DISPOSAL_RESTORE_PREVIOUS	3

#define PGIMGREADER_TRANSP_AUTO		1
#define PGIMGREADER_TRANSP_DISABLE	2
#define PGIMGREADER_TRANSP_ENABLE	3

class PGImageReader {
public:
	#ifdef PGWIN
		char* m_hData;
		char* m_hMask;
		int m_iImageRowBytes;
		int m_iMaskRowBytes;
		BITMAPINFO** m_hBitmapHeader;
	#endif

	#ifdef PGGTK
		GdkRgbCmap** m_hColorMap;		
		guint32 *m_hCMapData;
                gint m_iNbColors, m_iCMapColors;
		guchar* m_hData;
//		guchar* m_hMask;
		GdkBitmap* m_hMask;
		GdkGC* m_hMaskGC;
		int m_iImageRowBytes;
	#endif
public:
	int m_iBackgroundIndex;
	PGBYTE m_iNbPal, m_iCrPal, m_iBitsPerPel, m_bTransp;
	PGColor* m_hBackgroundColor;
	PGImage* m_hImage;	
public:
	PGImageReader(PGImage* image = null, PGColor* bkgcolor = null);
	virtual ~PGImageReader();
	void setImage(PGImage* image);
	void setImageSize(int rows, int cols);
	void setBackgroundColor(PGColor* bkgcolor);
	void setBackgroundIndex(int idx);
	void selectPalette(int pal);
	void setPalette(int ncolors);
	void setColor(int idx, int red, int green, int blue);
	void closePalette();
	const PGColor* getBackground();
	boolean isTransparent();
	void put(unsigned char *data, int row);
	// Realize the image
	void realizeImage();	
	// Is this an animated image reader?
	virtual boolean isAnimated();
	// Set number of loops
	virtual void setLoops(int loops);
	// Goto the next image
	virtual void nextImage(unsigned long offset, PGWORD xpos, PGWORD ypos, PGWORD delay, PGBYTE disposal);
	// Initialize this image reader
	virtual void initReader(PGBYTE transp, int maxrows, int maxcols, PGBYTE nbits, PGBYTE nbpal);	
private:
	void specificInitialize(boolean trans, int maxrows, int maxcols, PGBYTE nbpal);
	void specificImageSize(int width, int height);
	void specificConstruct();
};

class PGImage : public PGObject {
public:	
	#ifdef PGWIN
		HBITMAP m_hBitmap;
		HBITMAP m_hMask;
	#endif

	#ifdef PGGTK
		GdkPixmap* m_hPixmap;
		GdkBitmap* m_hMask;
	#endif
public:
	int m_iRows, m_iCols;
	PGBYTE m_iError;
	PGGraphics* m_hGraphics;
	PGWindow* m_hOwner;
public:
	// Constructor / Destructor
	PGImage();
	PGImage(PGWindow* owner);
	PGImage(const char* file, PGWindow* owner);
	PGImage(const char* file, PGColor* background, PGWindow* owner);
	virtual ~PGImage();
	// Graphics functions
	BOOL hasGraphics();
	PGGraphics* getGraphics();
	void removeGraphics();	
	// Init/Term functions
	void reinitialize();
	void init();
	void term();
	void init(const char* file, PGColor* background, PGWindow* owner);	
	// Error?
	BOOL hasError();
protected:	
	void initspecific();
	void termspecific();
};

class PGAnimatedImageFrame : public PGImage {
public:
	PGWORD m_iXPos, m_iYPos;
	PGBYTE m_iDisposalMethod;   
	PGWORD m_iDelay;
public:
	PGAnimatedImageFrame(PGWindow* owner);
	PGAnimatedImageFrame(const char* file, PGWindow* owner);
};

class PGAnimatedImage : public PGObject {
public:
	PGWindow* m_hOwner;
	int m_iRows, m_iCols, m_iLoops;
	PGBYTE m_iError;
	PGVector m_hImages;
public:
	// Constructor / Destructor
	PGAnimatedImage(PGWindow* owner);	
	PGAnimatedImage(const char* file, PGWindow* owner);
	virtual ~PGAnimatedImage();
	// Graphics functions
	BOOL hasError();
	int getNbFrames();
	PGVector* getFrames();
	PGAnimatedImageFrame* getFrame(int idx);	
	void init(const char* file, PGWindow* owner);
};

class PGAnimatedImageReader : public PGImageReader {
public:
	PGAnimatedImage* m_hAnimatedImage;
public:
	PGAnimatedImageReader(PGAnimatedImage* image);
	// Is this an animated image reader?
	virtual boolean isAnimated();
	// Set number of loops
	virtual void setLoops(int loops);
	// Goto the next image
	virtual void nextImage(unsigned long offset, PGWORD xpos, PGWORD ypos, PGWORD delay, PGBYTE disposal);
	virtual void initReader(PGBYTE transp, int maxrows, int maxcols, PGBYTE nbits, PGBYTE nbpal);	
};

class PGIcon : public PGObject {
public:
	#ifdef PGWIN
		HICON m_hWIcon;
	#endif
public:
	PGImage* m_hImage;
public:
	PGIcon(PGImage* image);
	virtual ~PGIcon();
	void init();
	void term();
};

#endif
