/***
 * Portable GUI
 * Code for PGImage
 * Copyright (C) 2000-2000 Jan Struyf <jan.struyf@cs.kuleuven.ac.be>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <portablegui.h>
#include <pgdraw/pgbmp.h>
#include <pgdraw/pggif.h>

#ifdef PGGCC
	#include <ctype.h>	// for toupper()
#endif

PGImageReader::PGImageReader(PGImage* image, PGColor* bkgcolor) {
	m_hImage = image;
	m_hBackgroundColor = bkgcolor;
	m_iBackgroundIndex = -1;
	m_iCrPal = 0;
	m_iBitsPerPel = 0;
	specificConstruct();
}

void PGImageReader::setImage(PGImage* image) {
	m_hImage = image;
}

void PGImageReader::setImageSize(int rows, int cols) {
	m_hImage->m_iRows = rows;	
	m_hImage->m_iCols = cols;
	m_bTransp = isTransparent() ? 1 : 0;
	specificImageSize(rows, cols);
}

void PGImageReader::setBackgroundColor(PGColor* bkgcolor) {
	m_hBackgroundColor = bkgcolor;
}

void PGImageReader::setBackgroundIndex(int idx) {
	m_iBackgroundIndex = idx;
}

void PGImageReader::selectPalette(int pal) {
	m_iCrPal = pal;
}

const PGColor* PGImageReader::getBackground() {
	return m_hBackgroundColor;	
}

void PGImageReader::initReader(PGBYTE transp, int maxrows, int maxcols, PGBYTE nbits, PGBYTE nbpal) {
	boolean mytrans;
	m_iBitsPerPel = nbits;
	m_iNbPal = nbpal;
	switch (transp) {
		case PGIMGREADER_TRANSP_AUTO:
			mytrans = isTransparent();
			break;
		case PGIMGREADER_TRANSP_DISABLE:
			mytrans = false;
			break;
		case PGIMGREADER_TRANSP_ENABLE:
			mytrans = true;
			break;
	}		
	specificInitialize(mytrans, maxrows, maxcols, nbpal);
}

boolean PGImageReader::isTransparent() {
	return (m_hBackgroundColor != null) || (m_iBackgroundIndex != -1);
}

boolean PGImageReader::isAnimated() {
	return false;
}

void PGImageReader::setLoops(int loops) {
}

void PGImageReader::nextImage(unsigned long offset, PGWORD xpos, PGWORD ypos, PGWORD delay, PGBYTE disposal) {
}

PGAnimatedImageReader::PGAnimatedImageReader(PGAnimatedImage* image) {
	m_hAnimatedImage = image;
}

boolean PGAnimatedImageReader::isAnimated() {
	return true;
}

void PGAnimatedImageReader::setLoops(int loops) {
	m_hAnimatedImage->m_iLoops = loops;
}

void PGAnimatedImageReader::initReader(PGBYTE transp, int maxrows, int maxcols, PGBYTE nbits, PGBYTE nbpal) {
	PGImageReader::initReader(transp, maxrows, maxcols, nbits, nbpal);
	m_hAnimatedImage->m_iRows = maxrows;
	m_hAnimatedImage->m_iCols = maxcols;
}

void PGAnimatedImageReader::nextImage(unsigned long offset, PGWORD xpos, PGWORD ypos, PGWORD delay, PGBYTE disposal) {
	PGAnimatedImageFrame* image = new PGAnimatedImageFrame(m_hAnimatedImage->m_hOwner);
	image->m_iXPos = xpos;
	image->m_iYPos = ypos;
	image->m_iDisposalMethod = disposal;
	image->m_iDelay = delay;
	image->m_iError = PGIMAGE_ERROR_NONE;
	m_hAnimatedImage->m_hImages.addElement(image);
	m_hImage = image;
}

PGImage::PGImage(PGWindow* owner) {
	init(null, null, owner);
}

PGImage::PGImage(const char* bmpfile, PGWindow* owner) {
	init(bmpfile, NULL, owner);
}

PGImage::PGImage(const char* bmpfile, PGColor* background, PGWindow* owner) {	
	init(bmpfile, background, owner);	
}

PGImage::PGImage() {
	init(NULL, NULL, NULL);
}

PGImage::~PGImage() {
	term();
}

void PGImage::reinitialize() {
	term();
	init();
}

void PGImage::init() {
	init(NULL, NULL, NULL);	
}

void PGImage::term() {
	removeGraphics();
	termspecific();
}

BOOL PGImage::hasGraphics() {
	return m_hGraphics != NULL;
}

void PGImage::removeGraphics() {
	if (m_hGraphics != NULL) {		
		PGDECREF(m_hGraphics);
		m_hGraphics = (PGGraphics*)NULL;
	}
}

void PGImage::init(const char* bmpfile, PGColor* background, PGWindow* owner) {
	initspecific();	
	m_hOwner = owner;	
	m_iRows = 0;
	m_iCols = 0;
	m_hGraphics = (PGGraphics*)NULL;
	m_iError = PGIMAGE_ERROR_NOCREATE;		
#ifdef PGFILE	
	if (bmpfile != NULL) {
		PGFileInputStream file(bmpfile);		
		if (file.getError() == PGINPUTSTREAM_ERROR_NONE) {
			PGImageReader reader(this, background);
			int len = strlen(bmpfile);
			if (toupper(bmpfile[len-1]) == 'P') {
#ifdef PGBMP			
				m_iError = PGLoadBMP(&file, &reader);
#endif				
			} else {
#ifdef PGGIF			
				m_iError = PGLoadGIF(&file, &reader);
#endif				
			}
		} else {
			m_iError = PGIMAGE_ERROR_FILE;	
		}
	}
#endif
}

BOOL PGImage::hasError() {
	return m_iError != PGIMAGE_ERROR_NONE;
}

PGAnimatedImageFrame::PGAnimatedImageFrame(PGWindow* owner) {
	init(NULL, NULL, owner);
}

PGAnimatedImageFrame::PGAnimatedImageFrame(const char* file, PGWindow* owner) {
	init(file, NULL, owner);
}

PGAnimatedImage::PGAnimatedImage(const char* file, PGWindow* owner) {
	init(file, owner);
}

PGAnimatedImage::PGAnimatedImage(PGWindow* owner) {
	init(null, owner);
}

void PGAnimatedImage::init(const char* file, PGWindow* owner) {
	m_iRows = 0;
	m_iCols = 0;
	m_iLoops = 0;
	m_hOwner = owner;
	m_iError = PGIMAGE_ERROR_NOCREATE;
#ifdef PGFILE
#ifdef PGGIF
	if (file != NULL) {
		PGFileInputStream file(file);	
		if (file.getError() == PGINPUTSTREAM_ERROR_NONE) {
			PGAnimatedImageReader reader(this);
			m_iError = PGLoadGIF(&file, &reader);	
		} else {
			m_iError = PGIMAGE_ERROR_FILE;	
		}
	}
#endif
#endif
}

PGAnimatedImage::~PGAnimatedImage() {
}

BOOL PGAnimatedImage::hasError() {
	return m_iError != PGIMAGE_ERROR_NONE;
}

int PGAnimatedImage::getNbFrames() {
	return m_hImages.getSize();
}

PGVector* PGAnimatedImage::getFrames() {
	return &m_hImages;
}

PGAnimatedImageFrame* PGAnimatedImage::getFrame(int idx) {
	return (PGAnimatedImageFrame*)m_hImages.getElementAt(idx);
}

PGIcon::PGIcon(PGImage* image) {
	m_hImage = image;
	PGINCREF(image);
	init();
}

PGIcon::~PGIcon() {
	PGDECREF(m_hImage);
	term();
}
