/***
 * Portable GUI
 * Code for PGVector
 * Copyright (C) 2000-2000 Jan Struyf <jan.struyf@cs.kuleuven.ac.be>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <portablegui.h>
#include <stdio.h>

// Create new growable integer array
PGVector::PGVector() {
	init(1, 5);
}

// Create new growable integer array with initial size and grow term
PGVector::PGVector(long size, long grow) {
	init(size, grow);
}

// Delete growable integer array
PGVector::~PGVector() {
	deleteAll();
	if (m_iArrSize != 0) delete m_pArray;
}

// String representation of this vector
void PGVector::toString(PGString* str) {
	char result[PGSTRLEN];
	PGObject::toString(str);
	str->substring(8);
	sprintf(result, "PGVector[%ld]", getSize());
	str->prepend(result);
}

// Get size
long PGVector::getSize() {
	return m_iSize;
}

// Set grow
void PGVector::setGrow(long grow) {
	m_iGrow = grow;
}

// Get integer @ given position
void* PGVector::getElementAt(long idx) {
	if (idx < m_iSize) return m_pArray[idx];
	else return NULL;
}

// Add integer at end of array
void PGVector::addElement(void* element) {
	m_iSize++;
	grow();
	m_pArray[m_iSize-1] = element;
	if (m_bDoCleanUp) PGINCREF(element);
}

// Remove an element of the array
void PGVector::removeElement(void* element) {
	long idx = indexOfReference(element);
	if (idx != -1) removeElementAt(idx);
}

// Finds index of element using '==' on pointer
long PGVector::indexOfReference(void* element) {
	for (long pos = 0; pos < m_iSize; pos++)
			if (m_pArray[pos] == element) return pos;
	return -1;
}

// Set integer @ given position (may be >= getSize() - array grows)
void PGVector::setElementAt(void* element, long idx) {
	m_iSize = m_iSize > idx ? m_iSize : idx+1;
	grow();
    if (m_pArray[idx] != null && m_bDoCleanUp) PGDECREF(m_pArray[idx]);
	m_pArray[idx] = element;
	if (element != null && m_bDoCleanUp) PGINCREF(element);
}

// Remove elements from first to last
void PGVector::removeElements(long first, long last) {
	long pos;
	int nb_del = last-first+1;
	m_iSize -= nb_del;
	if (m_bDoCleanUp) {
		for (pos = first; pos <= last; pos++) {
			if (m_pArray[pos] != null) PGDECREF(m_pArray[pos]);
			m_pArray[pos] = null;
		}
	}
	for (pos = first; pos < m_iSize; pos++)
		m_pArray[pos] = m_pArray[pos+nb_del];
	for (pos = 0; pos < nb_del; pos++)
	    m_pArray[m_iSize+pos] = null;
	shrink();
}

// Remove integer at given position
void PGVector::removeElementAt(long idx) {
	removeElements(idx, idx);
}

// Empty array
void PGVector::removeAll() {
	if (m_iSize != 0) removeElements(0, m_iSize-1);
}

// Delete all objects (if not higher reference count)
void PGVector::deleteAll() {
	if (m_bDoCleanUp) {
		PGString str;
		for (long idx = 0; idx < m_iSize; idx++) {
			PGObject* obj = (PGObject*)m_pArray[idx];
			if (obj != NULL) {
/*
				obj->toString(&str);
				str.prepend("Remove from vector: ");
				PGConsolePrint(str.getValue());
*/
				PGDECREF(obj);
			}
		}
	}
}

// Shrink array to convenient size
void PGVector::shrink() {
	if (m_iSize < m_iArrSize - m_iGrow) {
		long nSize = m_iSize + m_iGrow;
		void **nArray = new PGVoidPtr[nSize];
		for (long idx = 0; idx < nSize; idx++)
			nArray[idx] = m_pArray[idx];
		delete m_pArray;
		m_pArray = nArray;
		m_iArrSize = nSize;
	}
}

// Grow array to convenient size
void PGVector::grow() {
	if (m_iSize > m_iArrSize) {
	    long nSize = m_iSize + m_iGrow;
		void **nArray = new PGVoidPtr[nSize];
		for (long i = 0; i < m_iArrSize; i++)
			nArray[i] = m_pArray[i];
		for (long j = m_iArrSize; j < nSize; j++)
			nArray[j] = NULL;
		delete m_pArray;
		m_pArray = nArray;
		m_iArrSize = nSize;
	}
}

// Initialize new vector
void PGVector::init(long size, long grow) {
	m_bDoCleanUp = TRUE;
	m_pArray = new PGVoidPtr[size];
	m_iArrSize = size;
	for (long idx = 0; idx < m_iArrSize; idx++) m_pArray[idx] = NULL;
	m_iSize = 0;
	m_iGrow = grow;
}

void PGVector::setDoCleanUp(BOOL yesNo) {
	m_bDoCleanUp = yesNo;
}

boolean PGVector::empty() {
    return getSize() == 0;
}

void PGStack::push(void* element) {
    setElementAt(element, getSize());
}

void* PGStack::pop() {
    long pos = getSize()-1;
    void* element = getElementAt(pos);
    removeElementAt(pos);
    return element;
}
