/***
 * Portable GUI
 * Code for PGGraphics
 * Copyright (C) 2000-2000 Jan Struyf <jan.struyf@cs.kuleuven.ac.be>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <portablegui.h>

GdkPoint* PGGraphicsMakePointArray(int* xPoints, int* yPoints, int nPoints) {
	GdkPoint* points = new GdkPoint[nPoints];
	for (int ctr = 0; ctr < nPoints; ctr++) {
		points[ctr].x = xPoints[ctr];
		points[ctr].y = yPoints[ctr];
	}
	return points;
}

void PGGraphics::init() {
	m_hWidget = NULL;
	m_hDrawable = NULL;
	m_hFont	= NULL;
	m_hGC = NULL;
	m_hClip.x = 0;	
	m_hClip.y = 0;
}

void PGGraphics::realize() {	
	m_hGC = gdk_gc_new(m_hWidget->window);
}

void PGGraphics::unrealize() {	
	if (m_hGC != NULL) {
		gdk_gc_destroy(m_hGC);
		m_hGC = NULL;
	}
}

// Clears the specified rectangle by filling it with the background color. 
void PGGraphics::clearRect(int x, int y, int width, int height) {
	gdk_draw_rectangle(m_hDrawable, m_hGC, 1, x, y, width, height);
}

// Intersects the current clip with the specified rectangle. 
void PGGraphics::clipRect(int x, int y, int width, int height) {
	GdkRectangle clip, rect, dest;
	// Source rect is current clip
	clip.x = m_hClip.x; clip.y = m_hClip.y;
	clip.width = m_hClip.width; clip.height = m_hClip.height;
	// Second rect is supplied
	rect.x = x; rect.y = y;
	rect.width = width; rect.height = height;				
	gdk_rectangle_intersect(&clip, &rect, &dest);
	gdk_gc_set_clip_rectangle(m_hGC, &dest);
}

// Copies an area of the component by a distance specified by dx and dy.
void PGGraphics::copyArea(int x, int y, int width, int height, int dx, int dy) {
	gdk_draw_pixmap(m_hDrawable, m_hGC, m_hDrawable, x, y, dx, dy, width, height);
}
 
// Draws the outline of a circular or elliptical arc covering the specified rectangle.  
void PGGraphics::drawArc(int x, int y, int width, int height, int startAngle, int arcAngle) {	
	gint angle1 = (gint)startAngle*64;
	gint angle2 = (gint)arcAngle*64;
	gdk_draw_arc(m_hDrawable, m_hGC, 0, x, y, width, height, angle1, angle2);
}

// Draws a part of a given image.           
void PGGraphics::drawImage(PGImage* image, int dx, int dy, int sx, int sy, int width, int height) {
	if (image->m_hMask != NULL) {
		gdk_gc_set_clip_origin(m_hGC, dx, dy);
		gdk_gc_set_clip_mask(m_hGC, image->m_hMask);
		gdk_draw_pixmap(m_hDrawable, m_hGC, image->m_hPixmap, sx, sy, dx, dy, width, height);		
		gdk_gc_set_clip_origin(m_hGC, 0, 0);		
		gdk_gc_set_clip_mask(m_hGC, null);		
	} else {
		gdk_draw_pixmap(m_hDrawable, m_hGC, image->m_hPixmap, sx, sy, dx, dy, width, height);
	}
}

// Draws the specified image, scaling it on the fly. 
// FIXME: Not implemented yet
void PGGraphics::drawImage(PGImage* img, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2) {
}

// Draws a line, using the current color, between the points (x1, y1) and (x2, y2)  
void PGGraphics::drawLine(int x1, int y1, int x2, int y2) {	
	gdk_draw_line(m_hDrawable, m_hGC, x1, y1, x2, y2);
}

// Draws the outline of an oval.           
void PGGraphics::drawOval(int x, int y, int width, int height) {
	gdk_draw_arc(m_hDrawable, m_hGC, 0, x, y, width, height, 0, (gint)360*64);
}

// Draws a closed polygon defined by arrays of x and y coordinates.  
void PGGraphics::drawPolygon(int* xPoints, int* yPoints, int nPoints) {
	GdkPoint* points = PGGraphicsMakePointArray(xPoints, yPoints, nPoints);
	gdk_draw_polygon(m_hDrawable, m_hGC, 0, points, nPoints);
	delete points;
}
 
// Draws a sequence of connected lines defined by arrays of x and y coordinates.    
void PGGraphics::drawPolyline(int* xPoints, int* yPoints, int nPoints) {
	GdkPoint* points = PGGraphicsMakePointArray(xPoints, yPoints, nPoints);
	gdk_draw_lines(m_hDrawable, m_hGC, points, nPoints);  
	delete points;
}

// Draws the outline of the specified rectangle.  
void PGGraphics::drawRect(int x, int y, int width, int height) {
	gdk_draw_rectangle(m_hDrawable, m_hGC, 0, x, y, width, height);
}

// Draws an outlined round-cornered rectangle using this graphics context's current color.  
void PGGraphics::drawRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight) {
	int midw = width-arcWidth;
	int midh = height-arcHeight;	
	int arw = arcWidth/2;
	int arh = arcHeight/2;
	gdk_draw_line(m_hDrawable, m_hGC, x+arw, y, x+arw+midw, y);			// Top
	gdk_draw_line(m_hDrawable, m_hGC, x+arw, y+height, x+arw+midw, y+height);	// Bottom
	gdk_draw_line(m_hDrawable, m_hGC, x, y+arh, x, y+arh+midh);			// Left
	gdk_draw_line(m_hDrawable, m_hGC, x+width, y+arh, x+width, y+arh+midh);		// Right	
	gdk_draw_arc(m_hDrawable, m_hGC, 0, x+width-arcWidth, y, arcWidth, arcHeight, 0, (gint)90*64);					// UpperRight	
	gdk_draw_arc(m_hDrawable, m_hGC, 0, x, y, arcWidth, arcHeight, (gint)90*64, (gint)90*64);					// UpperLeft	
	gdk_draw_arc(m_hDrawable, m_hGC, 0, x, y+height-arcHeight, arcWidth, arcHeight, (gint)180*64, (gint)90*64);			// LowerLeft
	gdk_draw_arc(m_hDrawable, m_hGC, 0, x+width-arcWidth, y+height-arcHeight, arcWidth, arcHeight, (gint)270*64, (gint)90*64);	// LowerRight
}

// Draws the text given by the specified string, using this graphics context's current font and color.   
void PGGraphics::drawString(const char *str, int x, int y) {
	gint ascent;
	gdk_string_extents(m_hFont->m_hGTKFont, (const gchar*)str, null, null, null, &ascent, null);
	gdk_draw_string(m_hDrawable, m_hFont->m_hGTKFont, m_hGC, x, y+ascent, (const gchar*)str);
}

// Draws the text given by the specified string, using this graphics context's current font and color.   
void PGGraphics::drawString(PGString *str, int x, int y) {
	drawString(str->getValue(), x, y);
}
	
// Paints a 3-D highlighted rectangle filled with the current color.    
void PGGraphics::fill3DRect(int x, int y, int width, int height, BOOL raised) {
}

// Fills a circular or elliptical arc covering the specified rectangle.   
void PGGraphics::fillArc(int x, int y, int width, int height, int startAngle, int arcAngle) {
	gint angle1 = (gint)startAngle*64;
	gint angle2 = (gint)arcAngle*64;
	gdk_draw_arc(m_hDrawable, m_hGC, 1, x, y, width, height, angle1, angle2);
}

// Fills an oval bounded by the specified rectangle with the current color.  
void PGGraphics::fillOval(int x, int y, int width, int height) {
	gdk_draw_arc(m_hDrawable, m_hGC, 1, x, y, width, height, 0, (gint)360*64);
}

#ifdef PGSHAPE
	// Draws the outline of a polygon defined by the specified Polygon object.   
	void PGGraphics::drawPolygon(PGPolygon* p) {	
		p->realize();
		gdk_draw_polygon(m_hDrawable, m_hGC, 0, p->m_hGTKPoints, p->getNbPoints());
	}

	// Fills the polygon defined by the specified Polygon object.  
	void PGGraphics::fillPolygon(PGPolygon* p) {
		p->realize();
		gdk_draw_polygon(m_hDrawable, m_hGC, 1, p->m_hGTKPoints, p->getNbPoints());		
	}
#endif

// Fills a closed polygon defined by arrays of x and y coordinates.   
void PGGraphics::fillPolygon(int* xPoints, int* yPoints, int nPoints) {
	GdkPoint* points = PGGraphicsMakePointArray(xPoints, yPoints, nPoints);
	gdk_draw_polygon(m_hDrawable, m_hGC, 1, points, nPoints);
	delete points;
}

// Fills the specified rectangle.   
void PGGraphics::fillRect(int x, int y, int width, int height) {	
	gdk_draw_rectangle(m_hDrawable, m_hGC, 1, x, y, width, height);
}

// Fills the specified rounded corner rectangle with the current color.   
void PGGraphics::fillRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight) {
	int midw = width-arcWidth;
	int midh = height-arcHeight;	
	int arw = arcWidth/2;
	int arh = arcHeight/2;
	gdk_draw_rectangle(m_hDrawable, m_hGC, 1, x+arw, y, midw, height);	
	gdk_draw_rectangle(m_hDrawable, m_hGC, 1, x, y+arh, arw, midh); 								// Left
	gdk_draw_rectangle(m_hDrawable, m_hGC, 1, x+width-arw, y+arh, arw, midh);							// Right	
	gdk_draw_arc(m_hDrawable, m_hGC, 1, x+width-arcWidth, y, arcWidth, arcHeight, 0, (gint)90*64);					// UpperRight	
	gdk_draw_arc(m_hDrawable, m_hGC, 1, x, y, arcWidth, arcHeight, (gint)90*64, (gint)90*64);					// UpperLeft	
	gdk_draw_arc(m_hDrawable, m_hGC, 1, x, y+height-arcHeight, arcWidth, arcHeight, (gint)180*64, (gint)90*64);			// LowerLeft
	gdk_draw_arc(m_hDrawable, m_hGC, 1, x+width-arcWidth, y+height-arcHeight, arcWidth, arcHeight, (gint)270*64, (gint)90*64);	// LowerRight
}
 
// Gets the current clipping area.    
// Shape getClip();          

// Returns the bounding rectangle of the current clipping area. 
void PGGraphics::getClipBounds(PGRectangle* bounds) {
	*bounds = m_hClip;
}

// Gets this graphics context's current color.   
PGColor* PGGraphics::getColor() {
	return &m_Color;
}

// Returns true if the specified rectangular area intersects the bounding rectangle of the current clipping area.  
BOOL PGGraphics::hitClip(int x, int y, int width, int height) {
	GdkRectangle clip, rect;
	// Source rect is current clip
	clip.x = m_hClip.x; clip.y = m_hClip.y;
	clip.width = m_hClip.width; clip.height = m_hClip.height;
	// Second rect is supplied
	rect.x = x; rect.y = y;
	rect.width = width; rect.height = height;				
	return (gdk_rectangle_intersect(&clip, &rect, null) == 1);
}

// Sets the current clip to the rectangle specified by the given coordinates.   
void PGGraphics::setClip(int x, int y, int width, int height) {
	GdkRectangle rect;
	rect.x = m_hClip.x = x;
	rect.y = m_hClip.y = y;	
	rect.width = m_hClip.width = width;
	rect.height = m_hClip.height = height;
	gdk_gc_set_clip_rectangle(m_hGC, &rect);
}

// Sets this graphics context's current color to the specified color.    
void PGGraphics::setColor(PGColor* c) {
	if (!m_Color.equals(c)) {	
		m_Color.setCopyValue(c);
		GdkColor color;
		guint32 val = ((guint32)c->m_iRed << 16) + 
			      ((guint32)c->m_iGreen << 8) + 
			       (guint32)c->m_iBlue;			        
		color.pixel = gdk_rgb_xpixel_from_rgb(val);
		gdk_gc_set_foreground(m_hGC, &color);
	}
}

// Set the width for drawing lines
void PGGraphics::setLineWidth(int wd) {
	if (wd != m_iLineWidth) {
		m_iLineWidth = wd;
	}
}
 	
// Sets this graphics context's font  
void PGGraphics::setFont(PGRealizedFont* font) {
	m_hFont = font;
}

