/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/.
 */

#include <portablegui.h>
#include <pgdep/gtk/pggtkcanvas.h>

static void pg_gtk_layobox_class_init(PGGtkLayoboxClass *klass);
static void pg_gtk_layobox_init(PGGtkLayobox *layobox);
static void pg_gtk_layobox_map(GtkWidget *widget);
static void pg_gtk_layobox_unmap(GtkWidget *widget);
static void pg_gtk_layobox_draw(GtkWidget *widget, GdkRectangle *area);
static gint pg_gtk_layobox_expose(GtkWidget *widget, GdkEventExpose *event);
static void pg_gtk_layobox_add(GtkContainer *container, GtkWidget *widget);
static void pg_gtk_layobox_remove(GtkContainer *container, GtkWidget *widget);
static void pg_gtk_layobox_forall(GtkContainer *container, gboolean include_internals, GtkCallback callback, gpointer callback_data);
static GtkType pg_gtk_layobox_child_type(GtkContainer *container);
static void pg_gtk_layobox_size_request(GtkWidget *widget, GtkRequisition *requisition);
static void pg_gtk_layobox_size_allocate(GtkWidget *widget, GtkAllocation  *allocation);
 
static GtkContainerClass *parent_class = NULL;

GtkType pg_gtk_layobox_get_type(void) {
  static GtkType layobox_type = 0;
  if (!layobox_type) {
      static const GtkTypeInfo layobox_info = {
	"PGGtkLayobox",
	sizeof (PGGtkLayobox),
	sizeof (PGGtkLayoboxClass),
	(GtkClassInitFunc) pg_gtk_layobox_class_init,
	(GtkObjectInitFunc) pg_gtk_layobox_init,
        NULL, NULL,
	(GtkClassInitFunc) NULL
      };
      layobox_type = gtk_type_unique(GTK_TYPE_CONTAINER, &layobox_info);
  }
  return layobox_type;
}

GtkWidget* pg_gtk_layobox_new() {
  PGGtkLayobox* box = (PGGtkLayobox*)gtk_type_new(pg_gtk_layobox_get_type());
  return GTK_WIDGET(box);
}

static void pg_gtk_layobox_class_init(PGGtkLayoboxClass *pClass) {
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;
  GtkContainerClass *container_class;

  object_class = (GtkObjectClass*)pClass;
  widget_class = (GtkWidgetClass*)pClass;
  container_class = (GtkContainerClass*)pClass;

  parent_class = (GtkContainerClass*)gtk_type_class(GTK_TYPE_CONTAINER);

  widget_class->map = pg_gtk_layobox_map;
  widget_class->unmap = pg_gtk_layobox_unmap;
  widget_class->draw = pg_gtk_layobox_draw;
  widget_class->expose_event = pg_gtk_layobox_expose;

  container_class->add = pg_gtk_layobox_add;
  container_class->remove = pg_gtk_layobox_remove;
  container_class->forall = pg_gtk_layobox_forall;
  container_class->child_type = pg_gtk_layobox_child_type;
 
  widget_class->size_request = pg_gtk_layobox_size_request;
  widget_class->size_allocate = pg_gtk_layobox_size_allocate;
}

static void pg_gtk_layobox_init(PGGtkLayobox *layobox) {
  GTK_WIDGET_SET_FLAGS (layobox, GTK_NO_WINDOW);
  layobox->children = NULL;
  layobox->pgcontainer = NULL;
}

static GtkType pg_gtk_layobox_child_type(GtkContainer *container) {
  return GTK_TYPE_WIDGET;
}

static void pg_gtk_layobox_add(GtkContainer *container, GtkWidget *child) {
  g_return_if_fail (container != NULL);
  g_return_if_fail (PG_GTK_IS_LAYOBOX(container));
  g_return_if_fail (child != NULL);
  g_return_if_fail (child->parent == NULL);
 
  PGGtkLayobox *layobox = PG_GTK_LAYOBOX(container);

  layobox->children = g_list_append(layobox->children, child);
  gtk_widget_set_parent(child, GTK_WIDGET(layobox));
 
  if (GTK_WIDGET_REALIZED(layobox)) gtk_widget_realize(child);

  if (GTK_WIDGET_VISIBLE(layobox) && GTK_WIDGET_VISIBLE(child)) {
      if (GTK_WIDGET_MAPPED(layobox)) gtk_widget_map (child);
      gtk_widget_queue_resize(child);
  }
}

static void pg_gtk_layobox_remove(GtkContainer *container, GtkWidget *widget) {
  g_return_if_fail (container != NULL);
  g_return_if_fail (PG_GTK_IS_LAYOBOX(container));
  g_return_if_fail (widget != NULL);
  PGGtkLayobox *layobox = PG_GTK_LAYOBOX(container);
  GList *children = layobox->children;

  while (children) {
      GtkWidget* child = (GtkWidget*)children->data;

      if (child == widget) {
	  gboolean was_visible = GTK_WIDGET_VISIBLE(widget);
	  gtk_widget_unparent(widget);
	  layobox->children = g_list_remove_link(layobox->children, children);
	  g_list_free(children);
	
	  // queue resize regardless of GTK_WIDGET_VISIBLE (container),
	  // since that's what is needed by toplevels.
	  if (was_visible) gtk_widget_queue_resize(GTK_WIDGET(container));
	  break;
      }
      children = children->next;
    }
}

static void pg_gtk_layobox_map(GtkWidget *widget) {
  g_return_if_fail (widget != NULL);
  g_return_if_fail (PG_GTK_IS_LAYOBOX(widget));
 
  PGGtkLayobox *layobox = PG_GTK_LAYOBOX(widget);
  GTK_WIDGET_SET_FLAGS(layobox, GTK_MAPPED);
  GList *children = layobox->children;
  while (children) {
      GtkWidget* child = (GtkWidget*)children->data;
      children = children->next;

      if (GTK_WIDGET_VISIBLE(child) && !GTK_WIDGET_MAPPED(child)) gtk_widget_map(child);
  }
}

static void pg_gtk_layobox_unmap(GtkWidget *widget) {
  g_return_if_fail (widget != NULL);
  g_return_if_fail (PG_GTK_IS_LAYOBOX(widget));

  PGGtkLayobox *layobox = PG_GTK_LAYOBOX(widget);
  GTK_WIDGET_UNSET_FLAGS(layobox, GTK_MAPPED);
  GList *children = layobox->children;
  while (children) {
      GtkWidget* child = (GtkWidget*)children->data;
      children = children->next;

      if (GTK_WIDGET_VISIBLE(child) && GTK_WIDGET_MAPPED(child)) gtk_widget_unmap(child);
   }
}

static void pg_gtk_layobox_draw (GtkWidget *widget, GdkRectangle *area) {
  GdkRectangle child_area;
  g_return_if_fail (widget != NULL);
  g_return_if_fail (PG_GTK_IS_LAYOBOX(widget));
 
  if (GTK_WIDGET_DRAWABLE(widget)) {
      PGGtkLayobox *layobox = PG_GTK_LAYOBOX(widget);
      GList *children = layobox->children;
      
      while (children) {
	  GtkWidget* child = (GtkWidget*)children->data;
	  children = children->next;
	
	  if (GTK_WIDGET_DRAWABLE(child) && gtk_widget_intersect(child, area, &child_area)) {
	  	gtk_widget_draw(child, &child_area);
	  }
      }
   }
}

static gint pg_gtk_layobox_expose(GtkWidget *widget, GdkEventExpose *event) {
  GdkEventExpose child_event;
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (PG_GTK_IS_LAYOBOX (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);  

  if (GTK_WIDGET_DRAWABLE(widget)) {  
      child_event = *event;
 
      PGGtkLayobox *layobox = PG_GTK_LAYOBOX(widget);
      GList *children = layobox->children;
      
      while (children) {
	  GtkWidget* child = (GtkWidget*)children->data;
	  children = children->next;

	  if (GTK_WIDGET_DRAWABLE(child) && GTK_WIDGET_NO_WINDOW(child) && gtk_widget_intersect(child, &event->area, &child_event.area))
	        gtk_widget_event(child, (GdkEvent*)&child_event);
      }
  }
  return FALSE;
}

static void pg_gtk_layobox_forall(GtkContainer *container, gboolean include_internals, GtkCallback callback, gpointer callback_data) {
  g_return_if_fail (container != NULL);
  g_return_if_fail (PG_GTK_IS_LAYOBOX (container));
  g_return_if_fail (callback != NULL);

  PGGtkLayobox *layobox = PG_GTK_LAYOBOX(container);
  GList *children = layobox->children;
 
  while (children) {
      GtkWidget* child = (GtkWidget*)children->data;
      children = children->next;

      (* callback) (child, callback_data);
  }
}

void pg_gtk_layobox_set_pgcontainer(PGGtkLayobox *layobox, PGContainer *container) {
  g_return_if_fail (container != NULL);
  g_return_if_fail (layobox != NULL);

  layobox->pgcontainer = container;
}

void pg_gtk_layobox_append(GtkContainer *container, GtkWidget *child) {
  pg_gtk_layobox_add(container, child);
}

void pg_gtk_layobox_delete(GtkContainer *container, GtkWidget *widget) {
  pg_gtk_layobox_remove(container, widget);
}

static void pg_gtk_layobox_size_request(GtkWidget *widget, GtkRequisition *requisition) {
  PGDimension size;
  g_return_if_fail (widget != NULL);
  g_return_if_fail (PG_GTK_IS_LAYOBOX (widget));
  g_return_if_fail (requisition != NULL);

  PGGtkLayobox *layobox = PG_GTK_LAYOBOX(widget);
  if (layobox->pgcontainer != NULL)
	  layobox->pgcontainer->getPreferredSize(&size, TRUE);

  requisition->width  = size.width + GTK_CONTAINER(layobox)->border_width * 2;
  requisition->height = size.height + GTK_CONTAINER(layobox)->border_width * 2;
}

static void pg_gtk_layobox_size_allocate(GtkWidget *widget, GtkAllocation *allocation) {
  PGRectangle rectangle;
  g_return_if_fail (widget != NULL);
  g_return_if_fail (PG_GTK_IS_LAYOBOX (widget));
  g_return_if_fail (allocation != NULL);

  widget->allocation = *allocation;

  PGGtkLayobox *layobox = PG_GTK_LAYOBOX(widget);
  int border = GTK_CONTAINER(layobox)->border_width;
 
  rectangle.x = allocation->x + border;
  rectangle.y = allocation->y + border;
  rectangle.width = allocation->width - 2*border;
  rectangle.height = allocation->height - 2*border;

  if (layobox->pgcontainer != NULL)
	  layobox->pgcontainer->setLayoutSize(&rectangle);
}
