/***
 * Portable GUI
 * Code for PGImage
 * Copyright (C) 2000-2000 Jan Struyf <jan.struyf@cs.kuleuven.ac.be>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <portablegui.h>
#include <pgdraw/pgbmp.h>

int PGImageReaderCalcBoundary(int size, int boundary) {
	size = (size + boundary - 1) / boundary;
	return size*boundary;
}

void PGImageReader::specificConstruct() {
	m_hData = NULL;
	m_hMask = NULL;
	m_hBitmapHeader = NULL;
}

PGImageReader::~PGImageReader() {
	if (m_hData != NULL) delete[] m_hData;	
	if (m_hMask != NULL) delete[] m_hMask;
	if (m_hBitmapHeader != NULL) {
		for (PGBYTE i = 0; i < m_iNbPal; i++) 
			if (m_hBitmapHeader[i] != null) 
				delete[] m_hBitmapHeader[i];
		delete[] m_hBitmapHeader;
	}
}

void PGImageReader::specificImageSize(int rows, int cols) {
	// Calculate bitmap data size
	m_iImageRowBytes = ((m_iBitsPerPel * cols) + 3) / 4;
	m_iImageRowBytes = (int) PGImageReaderCalcBoundary(m_iImageRowBytes/2, sizeof(LONG));
	long size = (long)m_iImageRowBytes*rows;
	for (int ctr = 0; ctr < size; ctr++) m_hData[ctr] = 0;
	if (isTransparent()) {
		// Calculate mask size and clear mask
		m_iMaskRowBytes = PGImageReaderCalcBoundary((cols+7)/8, sizeof(WORD));		
		long size = (long)m_iMaskRowBytes*rows;
		for (int ctr = 0; ctr < size; ctr++) m_hMask[ctr] = 0;		
	} else {
		// Image is not transparent
		m_iMaskRowBytes = 0;	
	}
}

void PGImageReader::specificInitialize(boolean transp, int maxrows, int maxcols, PGBYTE nbpal) {
	// Allocate buffer	
	int mxrowbytes = ((m_iBitsPerPel * maxcols) + 3) / 4;
	mxrowbytes = (int)PGImageReaderCalcBoundary(mxrowbytes/2, sizeof(LONG));
	// Allocate and zero buffer
	long size = (long)mxrowbytes*maxrows;
	m_hData = new char[size];	
	// Background buffer?
	if (transp) {
		m_iMaskRowBytes = PGImageReaderCalcBoundary((maxcols+7)/8, sizeof(WORD));
		size = (long)m_iMaskRowBytes*maxrows;
		m_hMask = new char[size];		
	}
	// Create palette array
	m_hBitmapHeader = new LPBITMAPINFO[nbpal];
	for (PGBYTE i = 0; i < nbpal; i++) 
		m_hBitmapHeader[i] = null;
}

void PGImageReader::setPalette(int ncolors) {
	// Delete old palette if already exists
	if (m_hBitmapHeader[m_iCrPal] != null) delete m_hBitmapHeader[m_iCrPal];
	// Allocate header
	int wincolors = PGCOLORCNT(m_iBitsPerPel);	
	long size = sizeof(BITMAPINFOHEADER) + (wincolors ? sizeof(RGBQUAD)*wincolors : 0);
	BITMAPINFO* info = (BITMAPINFO*)new char[size];		
	m_hBitmapHeader[m_iCrPal] = info;
	// Fill header struct
	BITMAPINFOHEADER* header;		
	header = (BITMAPINFOHEADER*)&info->bmiHeader;
	header->biSize = sizeof(BITMAPINFOHEADER);
	header->biPlanes = 1;
	header->biBitCount = m_iBitsPerPel;
	header->biCompression = BI_RGB;
	header->biXPelsPerMeter = 0;
	header->biYPelsPerMeter = 0;
	header->biClrUsed = 0;
	header->biClrImportant = 0;
}

void PGImageReader::setColor(int idx, int red, int green, int blue) {
	BITMAPINFO* info = m_hBitmapHeader[m_iCrPal];
	RGBQUAD* color = (RGBQUAD*)&info->bmiColors;
	color[idx].rgbBlue = blue;
	color[idx].rgbGreen = green;
	color[idx].rgbRed = red;
	color[idx].rgbReserved = 0;
}

void PGImageReader::closePalette() {
}

void PGImageReader::realizeImage() {		
	// Create the pixmap
	int cols = m_hImage->m_iCols;
	int rows = m_hImage->m_iRows;
	// Header?	
	BITMAPINFO* info = m_hBitmapHeader[m_iCrPal];
	BITMAPINFOHEADER* header = (BITMAPINFOHEADER*)&info->bmiHeader;
	header->biSizeImage = (long)m_iImageRowBytes*rows;	
	header->biWidth = cols;
	header->biHeight = rows;
	// Data		
	if (m_hData != NULL) {
		// Create bitmap and mask
		HDC dc = GetDC(m_hImage->m_hOwner->m_hWnd);
		m_hImage->m_hBitmap = 
			CreateDIBitmap(dc, header, CBM_INIT, (CONST VOID*)m_hData, info, DIB_RGB_COLORS);
		if (isTransparent()) 
			m_hImage->m_hMask = CreateBitmap(cols, rows, 1, 1, m_hMask);			
		ReleaseDC(m_hImage->m_hOwner->m_hWnd, dc);
	}
}

void PGImageReader::put(unsigned char *data, int row) {
	int rows = m_hImage->m_iRows;
	int cols = m_hImage->m_iCols;
	int bits = m_iBitsPerPel;
	char* dest = (char*)m_hData+(rows-row-1)*m_iImageRowBytes;
	char* mask = (char*)m_hMask;
	if (mask != NULL) mask += row*m_iMaskRowBytes;
	// Fill row for 4/8bits color depth
	for (int ctr = 0; ctr < cols; ctr++) {
		char color = data[ctr];					
		if (mask != NULL && color == (char)m_iBackgroundIndex) {
			mask[ctr/8] |= 1 << (7-(ctr%8));					
		} else {
			if (bits == 8) dest[ctr] = color;
			else dest[ctr/2] |= color << (1-ctr%2)*4;
		}
	}	
}

void PGImage::termspecific() {
	if (m_hBitmap != NULL) DeleteObject(m_hBitmap);
	if (m_hMask != NULL) DeleteObject(m_hMask);
}

void PGImage::initspecific() {
	m_hBitmap = NULL;
	m_hMask = NULL;
}

PGGraphics* PGImage::getGraphics() {		
	if (m_hGraphics == NULL) {				
		HDC hDC = CreateCompatibleDC(NULL);	
		HGDIOBJ hObj = SelectObject(hDC, m_hBitmap);
		if ((hObj == NULL) || (hObj == (HGDIOBJ)GDI_ERROR)) {
			DeleteDC(hDC);		
		} else {
			m_hGraphics = new PGGraphics();
			PGINCREF(m_hGraphics);			
			m_hGraphics->m_iOwnerMode = PGGRAPHICS_OWNER_BITMAP;
			m_hGraphics->m_hWnd = m_hOwner->m_hWnd;
			m_hGraphics->m_hDC = hDC;		
			m_hGraphics->realize();
		}			
	}
	return m_hGraphics;
}

void PGIcon::init() {
	ICONINFO ii;
        ii.fIcon    = TRUE;
        ii.xHotspot = 0;
        ii.yHotspot = 0;
        ii.hbmMask  = m_hImage->m_hMask;
        ii.hbmColor = m_hImage->m_hBitmap;
        m_hWIcon = CreateIconIndirect(&ii);
}

void PGIcon::term() {
	DeleteObject(m_hWIcon);
}
