/***
 * Portable GUI
 * Code for PGApplication
 * Copyright (C) 2000-2000 Jan Struyf <jan.struyf@cs.kuleuven.ac.be>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <portablegui.h>
#include <commctrl.h>

BOOL PGApplicationRegisterFrameClass(HINSTANCE hInst, HBRUSH brush);
LRESULT CALLBACK PGApplicationWindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

#ifndef PGTINY
	LRESULT CALLBACK PGToolTipHookProc(int nCode, WPARAM wParam, LPARAM lParam);
	LRESULT CALLBACK PGCallWndHookProc(int  nCode, WPARAM  wParam, LPARAM  lParam);
	BOOL PGApplicationRegisterCanvasClass(HINSTANCE hInst, HBRUSH brush);
	LRESULT CALLBACK PGCanvasWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
#endif

PGApplication::~PGApplication() {
#ifndef PGTINY
	if (m_hToolTipHook != NULL) UnhookWindowsHookEx(m_hToolTipHook);
	if (m_hCallWndHook != NULL) UnhookWindowsHookEx(m_hCallWndHook);
#endif
	// Delete background brush
	DeleteObject(m_hBkBrush);
}

void PGApplication::exit(int code) {
	PostQuitMessage(code);
}

void PGApplication::println(const char* strg) {
	DWORD res;
	if (m_bConsoleEnabled) {
		HANDLE handle = GetStdHandle(STD_OUTPUT_HANDLE);
		WriteConsole(handle, strg, strlen(strg), &res, 0);
		WriteConsole(handle, "\r\n", 2, &res, 0);
	}
}

void PGApplication::printf(const char* format, ...) {
	DWORD res;
	int delta = 0;
	if (m_bConsoleEnabled) {
		HANDLE handle = GetStdHandle(STD_OUTPUT_HANDLE);
		va_list ap;
		int size = 30;
		char* str = new char[size];
		while (true) {
			va_start(ap, format);
			int n = _vsnprintf(str, size, format, ap);
			va_end(ap);
			if (n > -1 && n < size) {
				WriteConsole(handle, str, strlen(str), &res, 0);
				WriteConsole(handle, "\r\n", 2, &res, 0);
				delete[] str;
				return;
			}
			delete[] str;
			if (n > -1) size = n+1+delta;
			else size = size*3/2;
			str = new char[size];
		}
	}
}

void PGApplication::setConsoleEnabled(BOOL enable) {
	if (m_bConsoleEnabled != enable) {
		if (enable) AllocConsole();
		else FreeConsole();
		m_bConsoleEnabled = enable;
	}
}

int PGApplication::run() {
    MSG msg;
    // Message loop
    while (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
    // Unregister window classes
    UnregisterClass("pgframewnd", m_hInst);
#ifndef PGTINY
    UnregisterClass("pgcanvaswnd", m_hInst);
#endif
    return(msg.wParam);
}

HBRUSH PGApplicationCreateBkBrush() {
	LOGBRUSH brush;
	brush.lbStyle = BS_SOLID;
	brush.lbColor = GetSysColor(COLOR_BTNFACE);
	brush.lbHatch = 0;
	return CreateBrushIndirect(&brush);
}

void PGApplication::init(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine) {
	// Set reference to myself
	setPGApplication(this);
	// Init user interface libs
	InitCommonControls();
	// Remember startup directory
	m_sStartDir.reset(PGFILENAMELEN);
	GetCurrentDirectory(PGFILENAMELEN, m_sStartDir.getBuffer());
	m_sStartDir.terminate(PGSTRING_FILE_SEP);
	// Remember my instance
	m_hInst = hInst;
	m_sCmdLine.setCopyValue(lpCmdLine);
	// Create background brush
	m_hBkBrush = PGApplicationCreateBkBrush();
	// Register window classes
	if (!hPrevInst) {
		if (!PGApplicationRegisterFrameClass(hInst, m_hBkBrush)) exit(-1);
	}
#ifndef PGTINY
	if (!hPrevInst) {
		if (!PGApplicationRegisterCanvasClass(hInst, m_hBkBrush)) exit(-1);
	}
	// Hook for tooltips and focus,..
	m_hToolTipHook = SetWindowsHookEx(WH_GETMESSAGE, (HOOKPROC)PGToolTipHookProc,
        			(HINSTANCE) NULL, GetCurrentThreadId());
	m_hCallWndHook = SetWindowsHookEx(WH_CALLWNDPROC, (HOOKPROC)PGCallWndHookProc,
        			(HINSTANCE) NULL, GetCurrentThreadId());
#endif
}

BOOL PGApplicationRegisterFrameClass(HINSTANCE hInst, HBRUSH brush) {
    	WNDCLASS frameClass;
	frameClass.style         = CS_DBLCLKS; // Or 0?
	frameClass.lpfnWndProc   = (WNDPROC) PGApplicationWindowProc;
	frameClass.cbClsExtra    = 0;
	frameClass.cbWndExtra    = 0;
	frameClass.hInstance     = hInst;
	frameClass.hIcon         = LoadIcon(hInst, MAKEINTRESOURCE(100));
	frameClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
	frameClass.hbrBackground = brush;
	frameClass.lpszMenuName  = NULL;
	frameClass.lpszClassName = (LPSTR) "pgframewnd";
	return RegisterClass(&frameClass);
}

#define RETURNDEFAULTCALL -1

LRESULT CALLBACK PGApplicationWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	int iReturnValue = RETURNDEFAULTCALL;
	PGWindow* myWindow = (PGWindow*)GetWindowLong(hWnd, GWL_USERDATA);
	if (myWindow != NULL) {
		switch (uMsg) {
		    case WM_SYSCOMMAND:
		        break;
			case WM_CREATE:
				iReturnValue = 0;
				break;
			case WM_DESTROY:
				myWindow->notifyWindowListeners(PGWINDOW_CLOSED);
				SetWindowLong(hWnd, GWL_USERDATA, (LONG)NULL);
				iReturnValue = 0;
				break;
#ifndef PGTINY
			case WM_COMMAND:
				getPGApplication()->m_vCommandList.onAction(LOWORD(wParam));
				break;
#endif
#ifdef PGTIMER
			case WM_TIMER:
				getPGApplication()->m_vOtherIDList.onTimer(LOWORD(wParam));
				break;
#endif
#ifdef PGMENU
			case WM_MENUSELECT: {
					UINT uItem = (UINT) LOWORD(wParam);   // menu item or pop-up menu index
					UINT fuFlags = (UINT) HIWORD(wParam); // menu flags
					HMENU hmenu = (HMENU) lParam;         // handle of menu clicked
					if (fuFlags == 0xFFFF && hmenu == NULL) {
						PGFrame* frame = (PGFrame*)myWindow;
						if (frame->getMenuBar() != null)
						    frame->getMenuBar()->onMenuClosing();
					} else {
						getPGApplication()->m_vCommandList.onSelect(LOWORD(wParam));
					}
				}
				break;
#endif
			case WM_ACTIVATE: {
					int fActive = LOWORD(wParam);
					BOOL fMinimized = (BOOL)HIWORD(wParam);
					if (fActive == WA_INACTIVE) myWindow->notifyWindowListeners(PGWINDOW_DEACTIVATED);
					else myWindow->notifyWindowListeners(PGWINDOW_ACTIVATED);
				}
				break;

			case WM_SIZE: {
					int fwSizeType = wParam;
					if (fwSizeType == SIZE_MAXIMIZED || fwSizeType == SIZE_MINIMIZED || fwSizeType == SIZE_RESTORED) {
						if (fwSizeType == SIZE_MINIMIZED) {
							myWindow->m_bIconified = TRUE;
							myWindow->notifyWindowListeners(PGWINDOW_ICONIFIED);
						} else if (myWindow->m_bIconified) {
							myWindow->m_bIconified = FALSE;
							myWindow->notifyWindowListeners(PGWINDOW_DEICONIFIED);
						}
						if (!myWindow->m_bFirstShow) {
							myWindow->m_bFirstShow = TRUE;
							myWindow->notifyWindowListeners(PGWINDOW_OPENED);
						}
						myWindow->doLayout();
					}
					iReturnValue = 0;
				}
				break;

			case WM_CLOSE: {
					myWindow->notifyWindowListeners(PGWINDOW_CLOSING);
					iReturnValue = 0;
				}
				break;
		}
	}
	if (iReturnValue != RETURNDEFAULTCALL) return iReturnValue;
	else return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

#ifndef PGTINY
	BOOL PGApplicationRegisterCanvasClass(HINSTANCE hInst, HBRUSH brush) {
    		WNDCLASS canvasClass;
		canvasClass.style         = CS_DBLCLKS; // Or 0?
		canvasClass.lpfnWndProc   = (WNDPROC) PGCanvasWindowProc;
		canvasClass.cbClsExtra    = 0;
		canvasClass.cbWndExtra    = 0;
		canvasClass.hInstance     = hInst;
		canvasClass.hIcon         = NULL;
		canvasClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
		canvasClass.hbrBackground = brush;
		canvasClass.lpszMenuName  = NULL;
		canvasClass.lpszClassName = (LPSTR) "pgcanvaswnd";
		return RegisterClass(&canvasClass);
	}

	LRESULT CALLBACK PGCanvasWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
		int iReturnValue = RETURNDEFAULTCALL;
		PGCanvas* comp = (PGCanvas*)GetWindowLong(hWnd, GWL_USERDATA);
		if (comp != NULL) {
			switch (uMsg) {
				case WM_PAINT: {
						PAINTSTRUCT ps;
						BeginPaint(hWnd, &ps);
						PGGraphics graph;
						graph.m_iOwnerMode = PGGRAPHICS_OWNER_SYSTEM;
						graph.m_hWnd = hWnd;
						graph.m_hDC = ps.hdc;
						graph.realize();
						comp->paint(&graph);
						graph.unrealize();
						EndPaint(hWnd, &ps);
						iReturnValue = 0;
  					}
  					break;
  				// FIXME: This does not work!
  				case WM_CAPTURECHANGED:
  					getPGApplication()->println("Exit rgn!!");
  					comp->mouseExit(0, 0);
  					iReturnValue = 0;
  					break;
				case WM_RBUTTONDOWN:
					comp->mouseDown(LOWORD(lParam), HIWORD(lParam), PGMOUSEEVENT_BUTTON2_MASK);
					iReturnValue = 0;
					break;
				case WM_RBUTTONUP:
					comp->mouseUp(LOWORD(lParam), HIWORD(lParam), PGMOUSEEVENT_BUTTON2_MASK);
					iReturnValue = 0;
					break;
				case WM_LBUTTONDOWN:
					comp->mouseDown(LOWORD(lParam), HIWORD(lParam), PGMOUSEEVENT_BUTTON1_MASK);
					iReturnValue = 0;
					break;
				case WM_LBUTTONUP:
					comp->mouseUp(LOWORD(lParam), HIWORD(lParam), PGMOUSEEVENT_BUTTON1_MASK);
					iReturnValue = 0;
					break;
				case WM_MBUTTONDOWN:
					comp->mouseDown(LOWORD(lParam), HIWORD(lParam), PGMOUSEEVENT_BUTTON3_MASK);
					iReturnValue = 0;
					break;
				case WM_MBUTTONUP:
					comp->mouseUp(LOWORD(lParam), HIWORD(lParam), PGMOUSEEVENT_BUTTON3_MASK);
					iReturnValue = 0;
					break;
				case WM_MOUSEMOVE:
					comp->mouseMove(LOWORD(lParam), HIWORD(lParam));
					iReturnValue = 0;
					break;
			}
		}
		if (iReturnValue != RETURNDEFAULTCALL) return iReturnValue;
		else return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}


	#define PGCOMMANDLIST_OFFSET 100

	PGCommandList::PGCommandList() {
		m_vCommandMap.setGrow(20);
		m_vCommandMap.setDoCleanUp(FALSE);
		m_iFirstFree = 0;
	}

	void PGCommandList::onAction(int id) {
		if (id >= PGCOMMANDLIST_OFFSET) {
			PGActionListenerParent* action = (PGActionListenerParent*)m_vCommandMap.getElementAt(id-PGCOMMANDLIST_OFFSET);
			if (action != NULL) action->onAction();
		}
	}

	void PGCommandList::onSelect(int id) {
		if (id >= PGCOMMANDLIST_OFFSET) {
			PGActionListenerParent* action = (PGActionListenerParent*)m_vCommandMap.getElementAt(id-PGCOMMANDLIST_OFFSET);
			if (action != NULL) action->onSelect();
		}
	}

	void PGCommandList::onTimer(int id) {
#ifdef PGTIMER
		if (id >= PGCOMMANDLIST_OFFSET) {
			PGTimer* timer = (PGTimer*)m_vCommandMap.getElementAt(id-PGCOMMANDLIST_OFFSET);
			if (timer != NULL) PGTimerTick(timer);
		}
#endif		
	}

	int PGCommandList::registerObject(PGObject* action) {
		while (m_vCommandMap.getElementAt(m_iFirstFree) != NULL) m_iFirstFree++;
		m_vCommandMap.setElementAt(action, m_iFirstFree);
		m_iFirstFree++;
		return m_iFirstFree+PGCOMMANDLIST_OFFSET-1;
	}

	int PGCommandList::registerCommand(PGActionListenerParent* action) {
		while (m_vCommandMap.getElementAt(m_iFirstFree) != NULL) m_iFirstFree++;
		m_vCommandMap.setElementAt(action, m_iFirstFree);
		m_iFirstFree++;
		return m_iFirstFree+PGCOMMANDLIST_OFFSET-1;
	}

	void PGCommandList::freeCommand(int id) {
		int len = (int)m_vCommandMap.getSize();
		id -= PGCOMMANDLIST_OFFSET;
		if (id == len-1) {
			int prev = len-2;
			while (prev >= 0 && m_vCommandMap.getElementAt(prev) == NULL) prev--;
			m_vCommandMap.removeElements(prev+1, len-1);
			m_iFirstFree = MIN2(m_iFirstFree, prev+1);
		} else {
			m_vCommandMap.setElementAt(NULL, id);
			m_iFirstFree = MIN2(m_iFirstFree, id);
		}
	}

	// Relay tooltip events
	LRESULT CALLBACK PGToolTipHookProc(int nCode, WPARAM wParam, LPARAM lParam) {
		PGApplication* app = getPGApplication();
		if (nCode >= 0) {
			MSG *lpmsg = (MSG *)lParam;
			switch (lpmsg->message) {
				case WM_MOUSEMOVE:
				case WM_LBUTTONDOWN:
				case WM_LBUTTONUP:
				case WM_RBUTTONDOWN:
				case WM_RBUTTONUP: {
					PGNormalComponent* comp = (PGNormalComponent*)GetWindowLong(lpmsg->hwnd, GWL_USERDATA);
					if (comp != NULL && PGCOMPONENT_TEST_TOOLTIP(comp)) {
						PGWindow* myWindow = comp->getWindow();
						if (myWindow != NULL) {
	                				MSG msg;
							msg.lParam = lpmsg->lParam;
							msg.wParam = lpmsg->wParam;
							msg.message = lpmsg->message;
                					msg.hwnd = lpmsg->hwnd;
                					SendMessage(myWindow->m_hWndToolTip, TTM_RELAYEVENT, 0, (LPARAM)(LPMSG)&msg);
                				}
                			}
            			}
            		}
		}
		return CallNextHookEx(app->m_hToolTipHook, nCode, wParam, lParam);
	}

	// SetFocus/LostFocus events
	LRESULT CALLBACK PGCallWndHookProc(int  nCode, WPARAM  wParam, LPARAM  lParam) {
		PGApplication* app = getPGApplication();
		if (nCode >= 0) {
#ifdef PGFOCUS
			CWPSTRUCT *cwp = (CWPSTRUCT*)lParam;
			switch (cwp->message) {
            			case WM_SETFOCUS: {
						PGComponent* comp = (PGComponent*)GetWindowLong(cwp->hwnd, GWL_USERDATA);
     						if (comp != NULL) comp->onFocusGained(comp);
					}
					break;
            			case WM_KILLFOCUS: {
            					PGComponent* comp = (PGComponent*)GetWindowLong(cwp->hwnd, GWL_USERDATA);
     						if (comp != NULL) comp->onFocusLost(comp);
					}
            				break;
            		}
#endif            		
		}
		return CallNextHookEx(app->m_hCallWndHook, nCode, wParam, lParam);
	}

#endif

